#!/bin/bash

#####################################################
# rerun.sh
#
#    This script extract every frame in a trajectory/dump file to a new data 
#    file, and begin a new LAMMPS simulation, starting from that data file.
#    (A new directory is also created to store the simulation output.)
#
#    This program is typically used to repeatedly run short-duration
#    (or zero-duration simulations), from each frame in a trajectory
#    in order to which print some thermodynamic output to the lammps.log file.
#    This way, the rerun.sh script allows users to re-analyze the
#    data from old simulations using new parameters ("coeffs") and/or fixes.
#
#    This program generates a lot of temporary directories and files.
#    Once you extract the output you need from each directory, 
#    you should probaly discard this output.
#
#####################################################
USAGE_EXAMPLE=$(cat <<EOF
Typical usage:
 
./rerun.sh $LAMPCOM new.in file.data [auxfile1 auxfile2 ...] < dump.lammpstrj

$LAMPCOM the command you normally use to run lammps.
  (For example, in ubuntu linux, LAMPCOM="lmp_ubuntu".)
new.in 
  This is a LAMMPS input script containing the commands you want to be executed
  repeatedly for every frame in the trajectory.  You can change the force-field
  parameters (coeffs) and other settings if you want to.  
  (NOTE: If the input script contains any "include" commands, the included files
      must be located in the same directory where the input script is located.)
file.data
  This file is a data file describing the topology and (new) force field
  parameters of the system.  Coordinates, dipole moments and box size
  will be read from the standard input, which is assumed to be a lammps 
  dump file (see below):
dump.lammpstrj
  A lammps dump file/trajectory is read from the standard input (using < or |). 
  For every frame, a new lammps simulation is started using the settings above.
OPTIONAL:
auxfile1 auxfile2...
  Depending on how you run lammps, some force-field styles and fixes require
  additional input files (for example "pair_style table" or "pair_style sw").
  rerun.sh needs to know about these files, so it can copy them to the temporary
  directory it creates to run each simulation.  If you need additional files to
  run lammps, list them after the data file (at the end of the argument list).
  NOTE: These files must be located in the same directory with the input script.
EOF
)
#####################################################


# Check to see if "dump2data.py" is installed and in the user's path.
# If not, look for it in the local directory.
if which dump2data.py > /dev/null; then 
    DUMP2DATACOM=dump2data.py
else
    DUMP2DATACOM=./dump2data.py
    if [ ! -s ./dump2data.py ]; then
        echo "Error: rerun.sh depends on dump2data.py. You must install"
        echo "       dump2data.py, or copy it to the working directory."
        exit 1
    fi
fi


if [ "$#" -lt 3 ]; then
    echo "Error: rerun.sh requires 3 arguments. $USAGE_EXAMPLE"
    exit 2
fi


LAMMPSCOMMAND="$1"
INPUTSCRIPT="$2"
DATAFILE="$3"
if [ ! -s "$INPUTSCRIPT" ]; then
    echo "Error: Unable to open input script: \"$INPUTSCRIPT\""
    exit 3
fi
if [ ! -s "$DATAFILE" ]; then
    echo "Error: Unable to open data file: \"$DATAFILE\""
    exit 4
fi
shift
shift
shift
# Any remaining arguments are the names of additional "auxilliary" files
# which lammps also needs to run the simulation (besides the input script(s)
# and the data file).
AUXFILES="$@"



# Now attempt to infer the atom_style from the $INPUTSCRIPT file.
# We need to know the atom_style before we attempt to read the data file.
ATOMSTYLE=`awk '{if ($1=="atom_style") {style=$2; for(i=3;i<=NF;i++){style=style" "$i}}} END{print style}' < "$INPUTSCRIPT"`
if [ "$ATOMSTYLE" == "" ]; then
    ATOMSTYLE="full" # If the atom_style was not specified, assume it is "full"
fi


# Replace the "read_data" or "read_restart" commands with a new "read_data"
# command which reads a new data file (which we will construct later).
awk 'BEGIN{found=0} {if (($1=="read_data")||($1=="read_restart")||($1=="read_dump")) {found=1; print "# rerun.sh_commented_out: \""$0"\""; print "read_data tmp.data"} else{print $0}} END{if (! found) exit(1)}' < $INPUTSCRIPT > tmp.in
if [ $? -ne 0 ]; then 
    echo "ERROR: \"$INPUTSCRIPT\" is lacking a read_data/read_restart/read_dump"
    echo "       statement."
    echo "Explanation:"
    echo "       rerun.sh modifies the input script replacing these commands"
    echo "       with a command to read a temporary data file.  However rerun.sh"
    echo "       searches for these commands in your script in order to know where to"
    echo "       make the modification.  You must insert a read_data or a read_restart"
    echo "       command at the appropriate place in your input script."
    echo "       (Ie., somewhere after the \"initialization\" phase where the atom_style"
    echo "       is defined).  The easiest solution to this problem is to find the"
    echo "       original input script you used to generate the trajectory and copy the"
    echo "       read_data or read_restart (or read_dump) command into the same place"
    echo "       in the the current script. (Make sure it is not commented out. It does"
    echo "       not matter if the data/restart file it refers to no longer exists.)"
    exit 5
fi



SUPPORTINGFILES=`awk '{if ($1=="include") {print $2}}' < $INPUTSCRIPT`





process_frame()
{
            # 1) Create a new directory in which to store the simulation 
            #    for the frame we just finished reading.
            if [ -d rerun_frame${framecount} ]; then
                rm -rf rerun_frame${framecount}
            fi
            mkdir rerun_frame${framecount}
            cp tmp.in rerun_frame${framecount}/
            for f in $SUPPORTINGFILES; do 
		#First, strip quotes from file name
		fname=`echo $f | tr -d '"'`
		echo "  copying file \"$fname\" to rerun_frame${framecount}/"
                cp -f "$fname" rerun_frame${framecount}/
            done

            for f in $AUXFILES; do 
		#First, strip quotes from file name
		fname=`echo $f | tr -d '"'`
		echo "  copying file \"$fname\" to rerun_frame${framecount}/"
                cp -f "$fname" rerun_frame${framecount}/
            done

            # 2) The current contents of the "tmp.dump.lammpstrj" file should
            # contain the previous frame.  In that case, convert this fragment
            # of the dump file to a data file (so later we can run LAMMPS on it).
            "$DUMP2DATACOM" -atomstyle $ATOMSTYLE "$DATAFILE" < tmp.dump.lammpstrj > rerun_frame${framecount}/tmp.data
            if [ $? -ne 0 ]; then
                exit 6
            fi

            cd rerun_frame${framecount}/
            echo "Entering directory rerun_frame${framecount}/"

            # 3) Now run LAMMPS on this new data file
            "$LAMMPSCOMMAND" -i tmp.in
            if [ $? -ne 0 ]; then
                exit 7
            fi

            cd ../

            # 4) Increment the framecount
            framecount=`expr $framecount + 1`

            # 5) --- Finally, erase frame.lammpstrj.tmp ---
            echo "" > tmp.dump.lammpstrj
}




#####  Main Loop  #####################################
##  Loop over every frame in the trajectory.         ##
##  For every frame, create a new directory, and     ##
##  run LAMMPS using that frame as a starting point. ##
#######################################################
rm -f tmp.dump.lammpstrj
framecount=1
while read line; do
    if [ "$line" == "ITEM: TIMESTEP" ]; then
        if [ -s tmp.dump.lammpstrj ]; then
            #Then we are reading in a new frame.

	    process_frame

        fi # if [ -s tmp.dump.lammpstrj ]; then
    fi #if [ "$line" == "ITEM: TIMESTEP" ]; then

    echo $line >> tmp.dump.lammpstrj
done

# Process the final frame in the trajectory in the same way that we did above:
process_frame

# Delete temporary files in the global directory
rm -f tmp.in tmp.dump.lammpstrj

